"""Base class for robots

This file is a part of Bookmarks database and Internet robot.

"""

__author__ = "Oleg Broytman <phd@phdru.name>"
__copyright__ = "Copyright (C) 2000-2024 PhiloSoft Design"
__license__ = "GNU GPL"

__all__ = ['robot_base']


from base64 import b64encode
from urllib.parse import urlsplit, urljoin
import asyncio
import sys
import socket
import time

from m_lib.md5wrapper import md5wrapper
from m_lib.net.www.util import parse_time

from bkmk_objects import __version__, Robot
from parse_html import parse_html
import bkmk_objects


# Fake headers to pretend this is a real browser
request_headers = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,'
    'image/avif,image/webp,image/png,image/svg+xml,*/*;q=0.8',
    'Accept-Language': 'ru,en-US;q=0.7,en;q=0.3',
    'Cache-Control': 'max-age=300',
    'Connection': 'close',
    'Referer': '/',
    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:120.0) '
                  'Gecko/20100101 Firefox/120.0',
    'X-User-Agent': 'bookmarks_db %s (Python %d.%d.%d) ' % (
        __version__, *sys.version_info[:3]),
}


_reloc_dict = {
  301: "perm1.",
  302: "temp2.",
  303: "temp3.",
  307: "temp7.",
  308: "temp8.",
  "html": "html"
}


def _get_error(e):
    if isinstance(e, str):
        return e

    else:
        s = []
        for i in e:
            s.append("'%s'" % str(i).replace('\n', "\\n"))
        return "(%s)" % ' '.join(s)


class robot_base(Robot):
    # Icon cache; maps URL to a tuple (content type, data)
    # or None if there is no icon.
    icons = {}

    # Pass proxy from the environment like this:
    # BKMK_ROBOT=curl,requests:proxy=http%3a//localhost%3a8080
    # BKMK_ROBOT=requests,curl:proxy=socks5h%3a//localhost%3a1080
    proxy = None

    # Store hosts for which we already know they require proxy...
    proxy_ok = set()
    # ...but aren't accessible even through proxy
    proxy_error = set()

    timeout = 60
    ftp_timeout = 60

    report_checked = True

    def __init__(self, *args, **kw):
        for attr in 'timeout', 'ftp_timeout':
            value = getattr(self, attr)
            if not isinstance(value, int):
                value = int(value)
                setattr(self, attr, value)
        Robot.__init__(self, *args, **kw)
        socket.setdefaulttimeout(self.timeout)
        request_headers['X-User-Agent'] += self.version_str()

    def version_str(self):
        return self.__class__.__name__

    def check_bookmark(self, bookmark):
        if self.report_checked:
            self.log('Checked: %s' % bookmark.href)
        return asyncio.run(self.check_bookmark_async(bookmark))

    async def check_bookmark_async(self, bookmark):
        try:
            self.start = int(time.time())
            bookmark.icon = None

            if bookmark.charset:
                req_headers = request_headers.copy()
                req_headers['Accept-Charset'] = bookmark.charset
            else:
                req_headers = request_headers

            error, http_status_code, redirect_to, resp_headers, content = \
                await self.get_url(bookmark.href, req_headers)

            if error is not None:
                bookmark.error = str(error)
                return

            if http_status_code and (http_status_code >= 400):
                bookmark.error = 'Error %d' % http_status_code
                return

            if http_status_code and (http_status_code >= 300):
                self.set_redirect(bookmark, http_status_code, redirect_to)
                return

            size = 0
            last_modified = None

            if resp_headers:
                try:
                    size = resp_headers["Content-Length"]
                except KeyError:
                    pass

                try:
                    last_modified = resp_headers["Last-Modified"]
                except KeyError:
                    pass

                if last_modified:
                    last_modified = parse_time(last_modified)

            if not size:  # Could be None from headers
                size = len(content)

            if last_modified:
                last_modified = str(int(last_modified))
            else:
                last_modified = bookmark.last_visit

            bookmark.size = size
            bookmark.last_modified = last_modified

            charset = None
            if resp_headers:
                try:
                    content_type = resp_headers["Content-Type"]
                    self.log("   Content-Type   : %s" % content_type)
                    if content_type is None:
                        if b'html' in content.lower():
                            content_type = 'text/html'
                        else:
                            content_type = 'text/plain'
                        self.log("   Set Content-Type to: %s"
                                 % content_type)
                    try:
                        # extract charset from
                        # "text/html; charset=UTF-8, foo; bar"
                        content_type, charset = content_type.split(';', 1)
                        content_type = content_type.strip()
                        charset = charset.split('=')[1].strip().split(',')[0]
                        self.log("   HTTP charset   : %s" % charset)
                    except (ValueError, IndexError):
                        charset = None
                        self.log("   no charset in Content-Type header")
                    is_html = False
                    for ctype in ("text/html", "application/xhtml+xml"):
                        if content_type.startswith(ctype):
                            is_html = True
                            break
                    content_stripped = content.strip()
                    if content_stripped and charset \
                            and isinstance(content_stripped, bytes):
                        try:
                            content_stripped = content_stripped.decode(
                                charset, 'replace')
                        except LookupError:
                            charset = None
                            self.log("   unknown charset "
                                     "in Content-Type header")
                    if content_stripped and is_html:
                        parser = parse_html(
                            content_stripped, charset, self.log)
                        if charset:
                            bookmark.charset = charset
                        elif parser and parser.meta_charset:
                            bookmark.charset = parser.meta_charset
                        if parser:
                            bookmark.real_title = parser.title
                            icon = parser.icon
                        else:
                            icon = None
                        if not icon:
                            icon = "/favicon.ico"
                        icon_url = icon_url_log = urljoin(bookmark.href, icon)
                        if icon_url_log.startswith('data:'):
                            icon_url_log = icon_url_log[:50] + '...'
                        self.log("   looking for icon at: %s" % icon_url_log)
                        if icon_url in self.icons:
                            if self.icons[icon_url]:
                                bookmark.icon_href = icon_url
                                content_type, bookmark.icon = \
                                    self.icons[icon_url]
                                self.log("   cached icon    : %s"
                                         % content_type)
                            else:
                                self.log("   cached icon    : no icon")
                        elif icon_url.startswith('data:'):
                            content_type, icon_data = \
                                icon_url[len('data:'):].split(',', 1)
                            bookmark.icon_href = bookmark.icon = icon_url
                            self.log("   got data icon  : %s" % content_type)
                            self.icons[icon_url] = (content_type, icon_url)
                        else:
                            try:
                                _icon_url = icon_url
                                for i in range(8):
                                    icon_error, \
                                        icon_status_code, icon_redirect_to, \
                                        icon_headers, icon_data = \
                                        await self.get_url(
                                            _icon_url, request_headers)
                                    if icon_error:
                                        raise IOError("No icon: " + icon_error)
                                        break
                                    if icon_status_code and (
                                            icon_status_code >= 300):
                                        _icon_url = icon_redirect_to
                                        self.log("   redirect to : %s"
                                                 % _icon_url)
                                    else:
                                        if icon_data is None:
                                            raise IOError("No icon")
                                        break
                                else:
                                    raise IOError("Too many redirects")
                            except:
                                etype, emsg, _ = sys.exc_info()
                                self.log("   no icon        : %s %s"
                                         % (etype, emsg))
                                etype = emsg = _ = None
                                self.icons[icon_url] = None
                            else:
                                content_type = icon_headers["Content-Type"]
                                if content_type and (
                                    content_type.startswith("application/")
                                    or content_type.startswith("image/")
                                    or content_type.startswith("text/plain")
                                ):
                                    bookmark.icon_href = icon_url
                                    self.log("   got icon       : %s"
                                             % content_type)
                                    if (
                                        content_type.startswith("application/")
                                        or content_type.startswith(
                                            "text/plain")
                                    ):
                                        self.log("   non-image content type,"
                                                 " assume x-icon")
                                        content_type = 'image/x-icon'
                                    if not isinstance(icon_data, bytes):
                                        icon_data = icon_data.encode('latin1')
                                    bookmark.icon = "data:%s;base64,%s" \
                                        % (content_type,
                                           b64encode(icon_data).decode('ascii'))
                                    self.icons[icon_url] = (
                                        content_type, bookmark.icon)
                                else:
                                    self.log("   no icon        : "
                                             "bad content type '%s'"
                                             % content_type
                                             )
                                    self.icons[icon_url] = None
                        if parser and parser.refresh:
                            refresh = parser.refresh
                            try:
                                url = refresh.split('=', 1)[1]
                            except IndexError:
                                url = "self"
                            try:
                                timeout = float(refresh.split(';')[0])
                            except (IndexError, ValueError):
                                self.set_redirect(bookmark, "html",
                                                  "Bad redirect to %s (%s)"
                                                  % (url, refresh)
                                                  )
                            else:
                                try:
                                    timeout = int(refresh.split(';')[0])
                                except ValueError:
                                    pass  # float timeout
                                self.set_redirect(bookmark, "html",
                                                  "%s (%s sec)"
                                                  % (url, timeout)
                                                  )
                    elif charset:
                        bookmark.charset = charset

                    if not content_stripped:
                        self.log("   empty response, no content")
                    if not is_html:
                        self.log("   not html")
                except KeyError as key:
                    self.log("   no header: %s" % key)

            md5 = md5wrapper()
            if bookmark.href.startswith("ftp://"):
                # Pass welcome message through MD5
                ftp_welcome = self.get_ftp_welcome()
                if not isinstance(ftp_welcome, bytes):
                    ftp_welcome = ftp_welcome.encode(charset or 'utf-8')
                md5.update(ftp_welcome)

            if isinstance(content, bytes):
                md5.update(content)
            else:
                md5.update(content.encode(charset or 'utf-8'))
            bookmark.md5 = str(md5)

        except EOFError:
            bookmark.error = "Unexpected EOF (FTP server closed connection)"
            self.log('   EOF: %s' % bookmark.error)

        except KeyboardInterrupt:
            self.log("Keyboard interrupt (^C)")
            return

        except socket.error as e:
            bookmark.error = _get_error(e)
            self.log(bookmark.error)

        except:
            tty_pbar = bkmk_objects.tty_pbar
            if tty_pbar:
                tty_pbar.erase()
            import traceback
            traceback.print_exc()
            bookmark.error = "Exception!"
            self.log('   Exception: %s' % bookmark.error)
            if tty_pbar:
                tty_pbar.redraw()

        finally:
            self.finish_check_url(bookmark)

    async def get_url(self, url, req_headers):
        split_results = urlsplit(url)
        url_proto = split_results.scheme
        url_host = split_results.hostname

        if url_host in self.proxy_error:
            return 'see prev. error', None, None, None, None

        use_proxy = False
        if self.proxy:
            split_proxy = urlsplit(self.proxy)
            proxy_proto = split_proxy.scheme
            # Do not route ftp requests via http(s) proxy;
            # socks5 proxies are ok.
            if (url_proto == 'ftp'):
                use_proxy = proxy_proto.startswith('socks5')
            else:  # http(s)
                use_proxy = True

        if use_proxy and url_host in self.proxy_ok:
            self.log('   Immediately trying with the proxy')
            error, http_status_code, resp_headers, content = \
                await self.get(url, req_headers, use_proxy=True)
        else:
            error, http_status_code, resp_headers, content = \
                await self.get(url, req_headers)
            if error is None or \
                    url_host.startswith('localhost') or \
                    url_host.startswith('127.'):
                use_proxy = False
            else:
                self.log('   Error          : %s' % error)
                if use_proxy:
                    if http_status_code == 404:
                        use_proxy = False
                    else:
                        self.log('   Retrying with the proxy...')
                        error, http_status_code, resp_headers, content = \
                            await self.get(url, req_headers, use_proxy=True)
                        if error is None:
                            self.proxy_ok.add(url_host)
        if (error is not None) or (
            http_status_code and (http_status_code >= 400)
        ):
            if use_proxy:
                self.log('   Proxy error    : %s' % error)
                if url_host not in self.proxy_ok:
                    self.proxy_error.add(url_host)
            return error, http_status_code, None, None, None
        if http_status_code and (http_status_code >= 300):
            if resp_headers:
                redirect_to = resp_headers['Location']
            else:
                redirect_to = 'UNKNOWN'
            return None, http_status_code, redirect_to, None, None
        return None, None, None, resp_headers, content

    def set_redirect(self, bookmark, errcode, newurl):
        bookmark.moved = moved = "(%s) to %s" % (_reloc_dict[errcode], newurl)
        try:
            moved.encode('ascii')
        except UnicodeEncodeError:
            try:
                moved = moved.encode(bookmark.charset)
            except (LookupError, TypeError, UnicodeEncodeError):
                moved = moved.encode('utf-8')
        self.log('   Moved: %s' % moved)

    def finish_check_url(self, bookmark):
        start = self.start
        bookmark.last_tested = str(start)
        now = int(time.time())
        bookmark.test_time = str(now - start)
