"""Robot based on curl_multi and concurrent.futures,
processes multiple URLs in parallel (multithreaded).

This file is a part of Bookmarks database and Internet robot.

"""

__author__ = "Oleg Broytman <phd@phdru.name>"
__copyright__ = "Copyright (C) 2024 PhiloSoft Design"
__license__ = "GNU GPL"

__all__ = ['robot_curl']


from queue import Queue
from time import sleep
from urllib.parse import urlsplit, parse_qsl, quote, quote_plus, urlencode
import concurrent.futures
import threading

import certifi
import pycurl

from Robots.base import robot_base
from Robots.concurrent_futures import concurrent_futures


class robot_curl(concurrent_futures):
    def __init__(self, *args, **kw):
        self.concurrent_class = concurrent.futures.ThreadPoolExecutor
        concurrent_futures.__init__(self, *args, **kw)
        #
        # Rename self.log, create one log_lines list per URL (thread)
        self.file_log = self.log
        del self.log
        self.local = threading.local()

        # Queue to pass requests to the main loop
        self.queue = Queue()

        # Main loop: run curl_multi
        self.executor.submit(self.curl_multi)

    def __getattr__(self, attr):
        if attr != 'log':
            raise AttributeError(attr)
        try:
            return self.local.log.append
        except AttributeError:
            return self.file_log

    def version_str(self):
        return str(pycurl.version)

    def check_bkmk_task(self):
        return self.check_bkmk_thread

    def check_bkmk_thread(self, bookmark):
        self.local.log = log = []
        robot_base.check_bookmark(self, bookmark)
        return bookmark, log

    async def get(self, url, req_headers, use_proxy=False):
        # Queue to wait results
        queue = Queue(maxsize=1)
        self.queue.put((url, req_headers, use_proxy, queue))
        return queue.get()

    def curl_multi(self):
        _curl_multi = pycurl.CurlMulti()
        _curl_multi.wrappers = {}

        while True:
            if self.queue.empty():
                pass
            else:
                request = self.queue.get_nowait()
                if request is None:  # Signal to stop
                    return
                url, req_headers, use_proxy, queue = request
                cw = CurlWrapper(self, url, req_headers, use_proxy)
                _curl_multi.wrappers[cw.curl] = (cw, queue)
                _curl_multi.add_handle(cw.curl)

            _curl_multi.select(1.0)
            while True:
                ret, num_handles = _curl_multi.perform()
                _, completed, failed = _curl_multi.info_read()
                for c in completed:
                    _curl_multi.remove_handle(c)
                    cw, queue = _curl_multi.wrappers.pop(c)
                    queue.put_nowait(
                        (None, cw.getinfo(pycurl.HTTP_CODE),
                         cw.resp_headers, cw.body)
                    )
                for c, errcode, errmsg in failed:
                    _curl_multi.remove_handle(c)
                    cw, queue = _curl_multi.wrappers.pop(c)
                    queue.put_nowait(('Error: %d %s' % (errcode, errmsg),
                                      None, None, None))
                if ret != pycurl.E_CALL_MULTI_PERFORM:
                    break
            if num_handles == 0:
                sleep(1)

    def wait(self):
        super(robot_curl, self).wait()
        if not self.bookmarks and not self.pending:
            self.queue.put(None)  # Signal to stop

    def get_ftp_welcome(self):
        return ''  # We don't store welcome message yet


class CurlWrapper:
    def __init__(self, robot, url, req_headers, use_proxy):
        req_headers = ['%s: %s' % (k, v) for k, v in req_headers.items()]

        self.curl = curl = pycurl.Curl()
        self.resp_headers = {}
        self.body = b''

        # Do not follow redirects
        curl.setopt(pycurl.FOLLOWLOCATION, 0)
        # Lower security settings - we need to get as musch as possible
        curl.setopt(pycurl.SSL_CIPHER_LIST, 'ALL:@SECLEVEL=1')
        curl.setopt(pycurl.SSL_VERIFYHOST, 0)
        curl.setopt(pycurl.SSL_VERIFYPEER, 0)
        curl.setopt(curl.CAINFO, certifi.where())
        # Set timeouts to avoid hanging too long
        if url.startswith('ftp://'):
            timeout = robot.ftp_timeout
        else:
            timeout = robot.timeout
        curl.setopt(pycurl.CONNECTTIMEOUT, timeout)
        curl.setopt(pycurl.TIMEOUT, timeout)
        # Parse Last-Modified
        curl.setopt(pycurl.OPT_FILETIME, 1)

        if use_proxy:
            curl.setopt(pycurl.PROXY, robot.proxy)

        # Set up a callback to capture the headers and the body
        curl.setopt(pycurl.HEADERFUNCTION, self.header_callback)
        curl.setopt(pycurl.WRITEFUNCTION, self.body_callback)

        curl.setopt(pycurl.HTTPGET, 1)
        curl.setopt(pycurl.HTTPHEADER, req_headers)
        try:
            url.encode('ascii')
        except UnicodeEncodeError:
            url = encode_url(url)
        curl.setopt(pycurl.URL, url)

    def __getattr__(self, attr):
        return getattr(self.curl, attr)

    def header_callback(self, data):
        for encoding in 'ascii', 'latin1', 'utf-8':
            try:
                data = data.decode(encoding)
            except UnicodeDecodeError:
                pass
            else:
                break
        else:
            print("Error decoding header:", data)
            return
        if ':' in data:
            key, value = data.split(':', 1)
            self.resp_headers[key.title()] = value.strip()

    def body_callback(self, data):
        self.body += data


def encode_url(url, encoding='latin1'):
    split_results = urlsplit(url)
    protocol, netloc, path, query, tag = split_results
    user = split_results.username
    password = split_results.password
    host = split_results.hostname
    port = split_results.port

    if query:
        qlist = []
        for name, value in parse_qsl(query):
            if not isinstance(name, bytes):
                name = name.encode(encoding)
                value = value.encode(encoding)
            qlist.append((name, value))

    url = protocol + "://"
    if user:
        url += quote(user.encode(encoding))
        if password:
            url += ':' + quote(password.encode(encoding))
        url += '@'
    if host:
        if isinstance(host, bytes):
            host = host.decode(encoding)
        url += host.encode('idna').decode('ascii')
        if port:
            url += ':%d' % port
    if path:
        if protocol == "file":
            url += quote(path)
        else:
            url += quote(path.encode(encoding))
    if query:
        url += '?' + urlencode(qlist)
    if tag:
        url += '#' + quote_plus(tag.encode(encoding))

    return url
