"""Robot based on requests

This file is a part of Bookmarks database and Internet robot.

"""

__author__ = "Oleg Broytman <phd@phdru.name>"
__copyright__ = "Copyright (C) 2024 PhiloSoft Design"
__license__ = "GNU GPL"

__all__ = ['robot_requests']


from urllib.parse import urlsplit
import ftplib
import socket
import warnings

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.ssl_ import create_urllib3_context
import requests
import urllib3

from Robots.base import robot_base


class robot_requests(robot_base):
    def version_str(self):
        return 'python-requests urllib3/%s' % urllib3.__version__

    async def get(self, url, req_headers, use_proxy=False):
        if url.startswith('ftp://'):
            error, welcome, body = _get_ftp(url, self.timeout)
            if error is not None:
                return error, None, None, None, None
            self.welcome = welcome
            return None, None, None, body

        if use_proxy:
            proxies = {'http': self.proxy, 'https': self.proxy}
        else:
            proxies = None

        s = requests.Session()
        s.mount('https://', AllCiphersAdapter())

        error = r = None
        try:
            r = s.get(url, headers=req_headers, timeout=self.timeout,
                      allow_redirects=False, proxies=proxies,
                      verify=False)
        except requests.RequestException as e:
            error = str(e)
            return error, None, None, None

        return None, r.status_code, r.headers, r.content

    def get_ftp_welcome(self):
        welcome = self.welcome
        self.welcome = ''
        return welcome


def _get_ftp(url, timeout):
    split_results = urlsplit(url)
    path = split_results.path or '/'
    user = split_results.username or 'anonymous'
    password = split_results.password or 'ftp'
    host = split_results.hostname
    port = split_results.port or 21

    ftp = ftplib.FTP()
    try:
        ftp.connect(host, port, timeout)
        ftp.login(user, password)
        if path != '/':
            ftp.cwd(path)
    except (socket.error, ftplib.Error) as e:
        return "Error: %s" % e, None, None

    files = []
    try:
        files = ftp.nlst()
    except ftplib.error_perm as e:
        if str(e).startswith("550 No files found"):
            return None, ftp.welcome, ''
        return "Error: %s" % e, None, None
    except ftplib.Error as e:
        return "Error: %s" % e, None, None

    return None, ftp.welcome, '\n'.join(files)


# See https://lukasa.co.uk/2017/02/Configuring_TLS_With_Requests/

class AllCiphersAdapter(HTTPAdapter):
    """
    A TransportAdapter that re-enables 3DES support in Requests.
    """
    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(cert_reqs=0,
                                         ciphers='ALL:@SECLEVEL=1')
        kwargs['ssl_context'] = context
        return super(AllCiphersAdapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(cert_reqs=0,
                                         ciphers='ALL:@SECLEVEL=1')
        kwargs['ssl_context'] = context
        return super(AllCiphersAdapter, self).proxy_manager_for(
            *args, **kwargs)


warnings.filterwarnings('ignore', 'Unverified HTTPS request is being made')
